@@ -25,16 +25,16 @@ class CosineScheduler(BaseLinearWarmupScheduler):
 class PolyScheduler(BaseLinearWarmupScheduler):
     r"""Poly LR Scheduler.
 
-    :param: poly_order: float. lr scheduler decreases with steps.
+    :param poly_order: float. lr scheduler decreases with steps.
     """
 
-    def __init__(self, poly_order: float = 0.5, **kwargs):
+    def __init__(self, optimizer, poly_order: float = 0.5, **kwargs):
         self.poly_order = poly_order
 
         if poly_order <= 0:
             raise ValueError(f'[-] poly_order must be positive. {poly_order}')
 
-        super().__init__(**kwargs)
+        super().__init__(optimizer, **kwargs)
 
     def _step(self) -> float:
         return self.min_lr + (self.max_lr - self.min_lr) * (self.step_t - self.warmup_steps) ** self.poly_order
