@@ -1,7 +1,6 @@
 import math
 
 import torch
-from torch.optim.optimizer import Optimizer
 
 from pytorch_optimizer.base.exception import NoSparseGradientError, ZeroParameterSizeError
 from pytorch_optimizer.base.optimizer import BaseOptimizer
@@ -9,7 +8,7 @@ from pytorch_optimizer.base.types import BETAS, CLOSURE, DEFAULTS, LOSS, PARAMET
 from pytorch_optimizer.optimizer.gc import centralize_gradient
 
 
-class Adai(Optimizer, BaseOptimizer):
+class Adai(BaseOptimizer):
     r"""Disentangling the Effects of Adaptive Learning Rate and Momentum.
 
     :param params: PARAMETERS. iterable of parameters to optimize or dicts defining parameter groups.
@@ -17,7 +16,8 @@ class Adai(Optimizer, BaseOptimizer):
     :param betas: BETAS. coefficients used for computing running averages of gradient and the squared hessian trace.
     :param weight_decay: float. weight decay (L2 penalty).
     :param weight_decouple: bool. the optimizer uses decoupled weight decay as in AdamW.
-    :param use_stable_weight_decay: bool. perform stable weight decay.
+    :param fixed_decay: bool. fix weight decay.
+    :param stable_weight_decay: bool. perform stable weight decay.
     :param dampening: float. dampening for momentum. where dampening < 1, it will show some adaptive-moment behavior.
     :param use_gc: bool. use gradient centralization.
     :param eps: float. term added to the denominator to improve numerical stability.
@@ -30,38 +30,32 @@ class Adai(Optimizer, BaseOptimizer):
         betas: BETAS = (0.1, 0.99),
         weight_decay: float = 0.0,
         weight_decouple: bool = False,
-        use_stable_weight_decay: bool = False,
+        fixed_decay: bool = False,
+        stable_weight_decay: bool = False,
         dampening: float = 1.0,
         use_gc: bool = False,
         eps: float = 1e-3,
+        **kwargs,
     ):
-        self.lr = lr
-        self.betas = betas
-        self.weight_decay = weight_decay
-        self.weight_decouple = weight_decouple
-        self.use_stable_weight_decay = use_stable_weight_decay
-        self.dampening = dampening
-        self.use_gc = use_gc
-        self.eps = eps
+        self.validate_learning_rate(lr)
+        self.validate_betas(betas)
+        self.validate_non_negative(weight_decay, 'weight_decay')
+        self.validate_non_negative(eps, 'eps')
 
-        self.validate_parameters()
+        self.use_gc = use_gc
 
         defaults: DEFAULTS = {
             'lr': lr,
             'betas': betas,
             'weight_decay': weight_decay,
+            'weight_decouple': weight_decouple,
+            'fixed_decay': fixed_decay,
+            'stable_weight_decay': stable_weight_decay,
             'dampening': dampening,
             'eps': eps,
         }
         super().__init__(params, defaults)
 
-    def validate_parameters(self):
-        self.validate_learning_rate(self.lr)
-        self.validate_betas(self.betas)
-        self.validate_weight_decay(self.weight_decay)
-        self.validate_epsilon(self.eps)
-
-    @property
     def __str__(self) -> str:
         return 'Adai'
 
@@ -94,7 +88,7 @@ class Adai(Optimizer, BaseOptimizer):
 
                 grad = p.grad
                 if grad.is_sparse:
-                    raise NoSparseGradientError(self.__str__)
+                    raise NoSparseGradientError(str(self))
 
                 param_size += p.numel()
 
@@ -109,15 +103,19 @@ class Adai(Optimizer, BaseOptimizer):
                 state['step'] += 1
 
                 if self.use_gc:
-                    grad = centralize_gradient(grad, gc_conv_only=False)
+                    centralize_gradient(grad, gc_conv_only=False)
 
-                bias_correction2 = 1.0 - beta2 ** state['step']
+                bias_correction2: float = self.debias(beta2, state['step'])
 
-                if not self.use_stable_weight_decay and group['weight_decay'] > 0.0:
-                    if self.weight_decouple:
-                        p.mul_(1.0 - group['lr'] * group['weight_decay'])
-                    else:
-                        grad.add_(p, alpha=group['weight_decay'])
+                if not group['stable_weight_decay'] and group['weight_decay'] > 0.0:
+                    self.apply_weight_decay(
+                        p=p,
+                        grad=grad,
+                        lr=group['lr'],
+                        weight_decay=group['weight_decay'],
+                        weight_decouple=group['weight_decouple'],
+                        fixed_decay=group['fixed_decay'],
+                    )
 
                 exp_avg_sq = state['exp_avg_sq']
                 exp_avg_sq.mul_(beta2).addcmul_(grad, grad, value=1.0 - beta2)
@@ -131,37 +129,41 @@ class Adai(Optimizer, BaseOptimizer):
 
         for group in self.param_groups:
             beta0, beta2 = group['betas']
-            beta0_dp = math.pow(beta0, 1.0 - group['dampening'])
+            beta0_dp: float = math.pow(beta0, 1.0 - group['dampening'])
             for p in group['params']:
                 if p.grad is None:
                     continue
 
                 grad = p.grad
+
                 state = self.state[p]
 
-                if self.use_stable_weight_decay and group['weight_decay'] > 0.0:
-                    if self.weight_decouple:
-                        p.mul_(1.0 - group['lr'] * group['weight_decay'])
-                    else:
-                        grad.add_(p, alpha=group['weight_decay'])
+                if group['stable_weight_decay'] and group['weight_decay'] > 0.0:
+                    self.apply_weight_decay(
+                        p=p,
+                        grad=grad,
+                        lr=group['lr'],
+                        weight_decay=group['weight_decay'],
+                        weight_decouple=group['weight_decouple'],
+                        fixed_decay=group['fixed_decay'],
+                    )
+
+                bias_correction2: float = self.debias(beta2, state['step'])
 
                 exp_avg, exp_avg_sq = state['exp_avg'], state['exp_avg_sq']
 
-                bias_correction2 = 1.0 - beta2 ** state['step']
-
                 exp_avg_sq_hat = exp_avg_sq / bias_correction2
                 beta1 = (
-                    1.0 - (exp_avg_sq_hat / exp_avg_sq_hat_mean).pow(1.0 / (3.0 - 2.0 * group['dampening'])).mul(beta0)
-                ).clamp(0.0, 1.0 - group['eps'])
-                beta3 = (1.0 - beta1).pow(group['dampening'])
+                    1.0
+                    - (exp_avg_sq_hat / exp_avg_sq_hat_mean).pow_(1.0 / (3.0 - 2.0 * group['dampening'])).mul_(beta0)
+                ).clamp_(0.0, 1.0 - group['eps'])
+                beta3 = (1.0 - beta1).pow_(group['dampening'])
 
                 beta1_prod = state['beta1_prod']
                 beta1_prod.mul_(beta1)
 
-                bias_correction1 = 1.0 - beta1_prod
-
                 exp_avg.mul_(beta1).addcmul_(beta3, grad)
-                exp_avg_hat = exp_avg.div(bias_correction1).mul(beta0_dp)
+                exp_avg_hat = exp_avg.div(1.0 - beta1_prod).mul_(beta0_dp)
 
                 p.add_(exp_avg_hat, alpha=-group['lr'])
 
