@@ -1,6 +1,6 @@
 import itertools
 from enum import IntEnum
-from typing import List, Tuple, Union
+from typing import List, Optional, Tuple, Union
 
 import numpy as np
 import torch
@@ -29,7 +29,7 @@ class Graft:
     def __init__(self, *args):
         pass
 
-    def add_statistics(self, grad: torch.Tensor, unused_beta2: float):
+    def add_statistics(self, grad: torch.Tensor, unused_beta2: float) -> None:
         r"""Add the statistics."""
         pass
 
@@ -37,7 +37,7 @@ class Graft:
         r"""Get preconditioned gradient."""
         return grad
 
-    def update_momentum(self, update: torch.Tensor, unused_beta1: float) -> torch.Tensor:  # noqa: ARG002
+    def update_momentum(self, update: torch.Tensor, unused_beta1: float) -> torch.Tensor:
         r"""Update momentum."""
         return update
 
@@ -47,7 +47,7 @@ class SGDGraft(Graft):
 
     def __init__(self, var: torch.Tensor):
         super().__init__(var)
-        self.momentum: torch.Tensor = torch.zeros_like(var, device=var.device)
+        self.momentum: torch.Tensor = torch.zeros_like(var)
 
     def update_momentum(self, update: torch.Tensor, beta1: float) -> torch.Tensor:
         r"""Update momentum."""
@@ -63,11 +63,11 @@ class SQRTNGraft(Graft):
 
     def precondition_gradient(self, grad: torch.Tensor) -> torch.Tensor:
         r"""Get preconditioned gradient."""
-        return torch.ones_like(grad) * torch.sign(grad)
+        return grad.sign()
 
 
 class AdaGradGraft(SGDGraft):
-    r"""Graft using Adagrad. Essentially an implementation of Adagrad with momentum.
+    r"""Graft using AdaGrad. Essentially an implementation of AdaGrad with momentum.
 
     :param var: torch.Tensor. variable.
     :param diagonal_eps: float. diagonal epsilon.
@@ -76,15 +76,15 @@ class AdaGradGraft(SGDGraft):
     def __init__(self, var: torch.Tensor, diagonal_eps: float):
         super().__init__(var)
         self.diagonal_eps = diagonal_eps
-        self.statistics: torch.Tensor = torch.zeros_like(var, device=var.device)
+        self.statistics: torch.Tensor = torch.zeros_like(var)
 
-    def add_statistics(self, grad: torch.Tensor, _):
+    def add_statistics(self, grad: torch.Tensor, _) -> None:
         r"""Add the statistics."""
         self.statistics.add_(grad.pow(2))
 
     def precondition_gradient(self, grad: torch.Tensor) -> torch.Tensor:
         r"""Get preconditioned gradient."""
-        return grad / (torch.sqrt(self.statistics) + self.diagonal_eps)
+        return grad.div(self.statistics.sqrt().add_(self.diagonal_eps))
 
 
 class RMSPropGraft(SGDGraft):
@@ -97,15 +97,15 @@ class RMSPropGraft(SGDGraft):
     def __init__(self, var: torch.Tensor, diagonal_eps: float):
         super().__init__(var)
         self.diagonal_eps = diagonal_eps
-        self.statistics: torch.Tensor = torch.zeros_like(var, device=var.device)
+        self.statistics: torch.Tensor = torch.zeros_like(var)
 
-    def add_statistics(self, grad: torch.Tensor, beta2: float):
+    def add_statistics(self, grad: torch.Tensor, beta2: float) -> None:
         r"""Add the statistics."""
         self.statistics.mul_(beta2).addcmul_(grad, grad, value=1.0 - beta2)
 
     def precondition_gradient(self, grad: torch.Tensor) -> torch.Tensor:
         r"""Get preconditioned gradient."""
-        return grad / (torch.sqrt(self.statistics) + self.diagonal_eps)
+        return grad.div(self.statistics.sqrt().add_(self.diagonal_eps))
 
 
 class BlockPartitioner:
@@ -115,54 +115,70 @@ class BlockPartitioner:
         so we effectively have 4 variables of size (1024, 512) each.
 
     :param var: torch.Tensor. tensor variable.
+    :param rank: int. rank.
     :param block_size: int. block size.
     :param pre_conditioner_type: int type of pre-conditioner.
     """
 
-    def __init__(self, var: torch.Tensor, block_size: int, pre_conditioner_type: int):
-        self.shape: List[int] = var.shape
+    def __init__(self, var: torch.Tensor, rank: int, block_size: int, pre_conditioner_type: int):
+        self.shape: torch.Size = var.shape
 
-        self.splits: List[Tuple[int, np.ndarray]] = []
-        self.split_sizes: List[Tuple[int, np.ndarray]] = []
+        self.splits: List[Tuple[int, torch.Tensor]] = []
+        self.split_sizes: List[Tuple[int, torch.Tensor]] = []
 
-        split_sizes: List[np.ndarray] = []
+        split_sizes: List[torch.Tensor] = []
 
         # We split var into smaller blocks. Here we store the metadata to make that split.
         for i, d in enumerate(self.shape):
-            if 0 < block_size < d:
-                # d - 1, otherwise split appends a 0-size array.
-                num_split: int = (d - 1) // block_size
-                indices = (np.arange(num_split, dtype=np.int32) + 1) * block_size
-                sizes = np.ones(num_split + 1, dtype=np.int32) * block_size
-                sizes[-1] = d - indices[-1]
-                self.splits.append((i, indices))
-                self.split_sizes.append((i, sizes))
-                split_sizes.append(sizes)
-            else:
-                split_sizes.append(np.array([d], dtype=np.int32))
+            if block_size <= 0 or block_size >= d:
+                split_sizes.append(torch.tensor([d], dtype=torch.int32))
+                continue
+
+            # d - 1, otherwise split appends a 0-size array.
+            num_split: int = (d - 1) // block_size
+            indices = (torch.arange(num_split, dtype=torch.int32) + 1) * block_size
+
+            sizes: torch.Tensor = torch.full((num_split + 1,), block_size, dtype=torch.int32)
+            sizes[-1] = d - indices[-1]
+
+            self.splits.append((i, indices))
+            self.split_sizes.append((i, sizes))
+            split_sizes.append(sizes)
 
         self.num_splits: int = len(split_sizes)
-        self.pre_conditioner_shapes: List[List[int]] = []
-        for t in itertools.product(*split_sizes):
-            if not (pre_conditioner_type == PreConditionerType.ALL or self.num_splits <= 1):
-                t = t[:-1]
-            self.pre_conditioner_shapes.extend([[d, d] for d in t])
+        self.pre_conditioner_shapes: List[List[torch.Tensor]] = self.build_pre_conditioner_shapes(
+            split_sizes, pre_conditioner_type, rank
+        )
 
-    def shapes_for_pre_conditioners(self) -> List[List[int]]:
+    @staticmethod
+    def build_pre_conditioner_shapes(
+        split_sizes: List[torch.Tensor], pre_conditioner_type: int, rank: int
+    ) -> List[List[torch.Tensor]]:
+        r"""Build pre-conditioner shapes."""
+        pre_conditioner_shapes: List[List[torch.Tensor]] = []
+        for t in itertools.product(*split_sizes):
+            t_shape: List[Optional[List[torch.Tensor]]] = [[d, d] for d in t]
+            if pre_conditioner_type == PreConditionerType.INPUT:
+                t_shape[-1] = None
+            elif pre_conditioner_type == PreConditionerType.OUTPUT:
+                t_shape = [None] * (rank - 1) + t_shape[-1:]
+            pre_conditioner_shapes.extend(t_shape)
+        return pre_conditioner_shapes
+
+    def shapes_for_pre_conditioners(self) -> List[List[torch.Tensor]]:
         r"""Get shapes of pre-conditioner."""
         return self.pre_conditioner_shapes
 
+    @torch.no_grad()
     def partition(self, x: torch.Tensor) -> List[torch.Tensor]:
         r"""Partition tensor into blocks."""
         if x.shape != self.shape:
-            raise ValueError(f'self._shape != x.shape ({self.shape} vs {x.shape})')
+            raise ValueError(f'self.shape != x.shape ({self.shape} vs {x.shape})')
 
-        tensors: List[torch.Tensor] = [x]
+        tensors = [x]
         for i, sizes in self.split_sizes:
-            tensors_local: List[torch.Tensor] = []
-            for t in tensors:
-                tensors_local.extend(torch.split(t, list(sizes), dim=i))
-            tensors = tensors_local
+            tensors = [torch.split(t, list(sizes), dim=i) for t in tensors]
+            tensors = [t for tensor in tensors for t in tensor]
         return tensors
 
     def merge_partitions(self, partitions: List[torch.Tensor]) -> torch.Tensor:
@@ -171,12 +187,9 @@ class BlockPartitioner:
             n: int = len(indices) + 1
 
             partitions: List[torch.Tensor] = [
-                torch.cat(partitions[idx:idx + n], axis=i) for idx in range(0, len(partitions), n)  # fmt: skip
+                torch.cat(partitions[idx:idx + n], dim=i) for idx in range(0, len(partitions), n)  # fmt: skip
             ]
 
-        # TODO
-        # when length of partitions is 1, raise error
-
         return partitions[0]
 
 
@@ -184,25 +197,27 @@ class PreConditionerType(IntEnum):
     r"""Type of PreConditioner.
 
     In default (ALL), computes pre-conditioner for each dim.
-    INPUT is one-sided Shampoo, in this case only on input dim.
+    INPUT/OUTPUT is one-sided Shampoo, in this case only on input/output dim.
     Assumes last dim is always the output dim and everything else input dim.
     """
 
     ALL = 0
     INPUT = 1
+    OUTPUT = 2
 
 
 class PreConditioner:
-    r"""Compute statistics/shape from gradients for preconditioning.
+    r"""Compute statistics & shape from gradients for preconditioning.
 
     :param var: torch.Tensor. variable.
     :param beta2: float. beta2.
-    :param inverse_exponent_override: int.
-    :param block_size: int.
-    :param no_preconditioning_for_layers_with_dim_gt: int.
-    :param shape_interpretation: bool.
-    :param matrix_eps: float.
+    :param inverse_exponent_override: int. override inv exp.
+    :param block_size: int. size of block.
+    :param skip_preconditioning_rank_lt: int. skip low-rank parameter.
+    :param no_preconditioning_for_layers_with_dim_gt: int. skip large size of dim of parameter.
+    :param shape_interpretation: bool. reshaping parameter.
     :param pre_conditioner_type: int. type of pre-conditioner.
+    :param matrix_eps: float. epsilon of matrix.
     :param use_svd: bool. use SVD instead of Schur-Newton method to calculate M^{-1/p}.
     """
 
@@ -212,31 +227,29 @@ class PreConditioner:
         beta2: float,
         inverse_exponent_override: int,
         block_size: int,
+        skip_preconditioning_rank_lt: int,
         no_preconditioning_for_layers_with_dim_gt: int,
         shape_interpretation: bool,
+        pre_conditioner_type: int,
         matrix_eps: float = 1e-6,
-        pre_conditioner_type: int = PreConditionerType.ALL,
         use_svd: bool = False,
     ):
         self.beta2 = beta2
         self.inverse_exponent_override = inverse_exponent_override
+        self.skip_preconditioning_rank_lt = skip_preconditioning_rank_lt
         self.no_preconditioning_for_layers_with_dim_gt = no_preconditioning_for_layers_with_dim_gt
-        self.matrix_eps = matrix_eps
         self.pre_conditioner_type = pre_conditioner_type
+        self.matrix_eps = matrix_eps
         self.use_svd = use_svd
 
         self.w2: float = 1.0 if self.beta2 == 1.0 else (1.0 - self.beta2)
 
-        self.original_shape: List[int] = var.shape
-        self.transformed_shape: List[int] = var.shape
-        if shape_interpretation:
-            self.transformed_shape = merge_small_dims(self.original_shape, block_size)
-
-        self.should_precondition_dims: List[bool] = (
-            [True] * len(self.transformed_shape)
-            if self.pre_conditioner_type == PreConditionerType.ALL or len(self.transformed_shape) <= 1
-            else [True] * (len(self.transformed_shape) - 1) + [False]
+        self.original_shape: torch.Size = var.shape
+        self.transformed_shape: List[int] = (
+            merge_small_dims(self.original_shape, block_size) if shape_interpretation else var.shape
         )
+
+        self.should_precondition_dims: List[bool] = self.get_should_precondition_dims()
         self.rank: int = sum(self.should_precondition_dims)
         self.exponent_for_pre_conditioner: int = (
             self.inverse_exponent_override if self.inverse_exponent_override > 0 else 2 * self.rank
@@ -244,24 +257,41 @@ class PreConditioner:
 
         self.statistics: Union[List[torch.Tensor], torch.Tensor] = []
         self.pre_conditioners: Union[List[torch.Tensor], torch.Tensor] = []
+
         self.is_same_shapes: bool = False
         if len(self.transformed_shape) > 1 and not self.skip_precondition(var):
-            reshaped_var = torch.reshape(var, self.transformed_shape)
-            self.partitioner = BlockPartitioner(reshaped_var, block_size, self.pre_conditioner_type)
+            self.partitioner = BlockPartitioner(
+                var=torch.reshape(var, self.transformed_shape),
+                rank=self.rank,
+                block_size=block_size,
+                pre_conditioner_type=self.pre_conditioner_type,
+            )
 
-            shapes = self.partitioner.shapes_for_pre_conditioners()
-            self.statistics = [self.matrix_eps * torch.eye(s[0], device=var.device) for s in shapes]
-            self.pre_conditioners = [torch.eye(s[0], device=var.device) for s in shapes]
-            self.is_same_shapes = len(np.unique(shapes)) == 1
+            shapes: List[Optional[List[torch.Tensor]]] = self.partitioner.shapes_for_pre_conditioners()
+            self.statistics = [self.matrix_eps * torch.eye(shape[0], device=var.device) for shape in shapes if shape]
+            self.pre_conditioners = [torch.eye(shape[0], device=var.device) for shape in shapes if shape]
+            self.is_same_shapes = None not in shapes and len(np.unique(shapes)) == 1
 
         if self.is_same_shapes:
-            self.statistics = torch.stack(self.statistics)
-            self.pre_conditioners = torch.stack(self.pre_conditioners)
+            self.statistics = torch.stack(self.statistics, dim=0)
+            self.pre_conditioners = torch.stack(self.pre_conditioners, dim=0)
+
+    def get_should_precondition_dims(self) -> List[bool]:
+        r"""Get pre-condition dimensions by the type of conditioner."""
+        if self.pre_conditioner_type == PreConditionerType.ALL or len(self.transformed_shape) <= 1:
+            return [True] * len(self.transformed_shape)
+        if self.pre_conditioner_type == PreConditionerType.INPUT:
+            return [True] * (len(self.transformed_shape) - 1) + [False]
+        if self.pre_conditioner_type == PreConditionerType.OUTPUT:
+            return [False] * (len(self.transformed_shape) - 1) + [True]
+        raise ValueError
 
     def skip_precondition(self, x: torch.Tensor) -> bool:
-        return (len(x.shape) < 1) or any([s > self.no_preconditioning_for_layers_with_dim_gt for s in x.shape])
+        return (len(x.shape) < self.skip_preconditioning_rank_lt) or any(
+            dim > self.no_preconditioning_for_layers_with_dim_gt for dim in x.shape
+        )
 
-    def add_statistics(self, grad: torch.Tensor):
+    def add_statistics(self, grad: torch.Tensor) -> None:
         r"""Compute statistics from gradients and add to the correct state entries.
 
         :param grad: torch.Tensor. gradient to compute statistics from.
@@ -272,33 +302,29 @@ class PreConditioner:
         reshaped_grad: torch.Tensor = torch.reshape(grad, self.transformed_shape)
         partitioned_grads: List[torch.Tensor] = self.partitioner.partition(reshaped_grad)
 
-        for j in range(len(partitioned_grads)):
-            partitioned_grad: torch.Tensor = partitioned_grads[j]
+        for j, partitioned_grad in enumerate(partitioned_grads):
             for i in range(self.rank):
                 axes: List[int] = [ax for ax in range(partitioned_grad.ndim) if ax != i]
                 stat: torch.Tensor = torch.tensordot(partitioned_grad, partitioned_grad, dims=[axes, axes])
                 self.statistics[j * self.rank + i].mul_(self.beta2).add_(stat, alpha=self.w2)
 
-    def compute_pre_conditioners(self):
+    def compute_pre_conditioners(self) -> None:
         r"""Compute L^{-1/exp} for each stats matrix L.
 
-        If `self.use_svd` is enabled,
-            where all shapes of statistics & pre-conditioners are same, perform batch SVD
-            else, SVD one by one.
-        else (`self.use_svd` is disabled), use Schur-Newton method
+        If `self.use_svd` is enabled and where all shapes of statistics & pre-conditioners are same, perform batch SVD.
+        else, SVD one by one.
+        If `self.use_svd` is disabled, use Schur-Newton method, which is usually much faster.
         """
         if self.use_svd and self.is_same_shapes:
-            self.pre_conditioners = compute_power_svd(
-                matrix=self.statistics, power=-1.0 / self.exponent_for_pre_conditioner
-            )
+            self.pre_conditioners = compute_power_svd(matrix=self.statistics, power=self.exponent_for_pre_conditioner)
             return
 
-        for i in range(len(self.statistics)):
+        for i, statistic in enumerate(self.statistics):
             self.pre_conditioners[i] = (
-                compute_power_svd(matrix=self.statistics[i], power=-1.0 / self.exponent_for_pre_conditioner)
+                compute_power_svd(matrix=statistic, power=self.exponent_for_pre_conditioner)
                 if self.use_svd
                 else compute_power_schur_newton(
-                    mat_g=self.statistics[i], p=self.exponent_for_pre_conditioner, ridge_epsilon=self.matrix_eps
+                    mat_g=statistic, p=self.exponent_for_pre_conditioner, ridge_epsilon=self.matrix_eps
                 )
             )
 
@@ -306,7 +332,7 @@ class PreConditioner:
     def precondition_block(
         partitioned_grad: torch.Tensor,
         should_preconditioned_dims: List[bool],
-        pre_conditioners_for_grad: List[torch.Tensor],
+        pre_conditioners_for_grad: Union[List[torch.Tensor], torch.Tensor],
     ) -> torch.Tensor:
         r"""Perform a preconditioning operation on a single gradient block.
 
@@ -314,12 +340,17 @@ class PreConditioner:
         We keep all axes in the same cyclic order they were originally.
         """
         rank: int = len(partitioned_grad.shape)
-        roll: Tuple[int, ...] = (*tuple(range(1, rank)), 0)
-        for j, should_precondition in enumerate(should_preconditioned_dims):
-            if not should_precondition:
-                partitioned_grad = torch.permute(partitioned_grad, roll).contiguous()
+        roll: Tuple[int, ...] = (*range(1, rank), 0)
+
+        i: int = 0
+        for should_precondition_dim in should_preconditioned_dims:
+            if not should_precondition_dim:
+                partitioned_grad = torch.permute(partitioned_grad, roll)
                 continue
-            partitioned_grad = torch.tensordot(partitioned_grad, pre_conditioners_for_grad[j], dims=[[0], [0]])
+
+            partitioned_grad = torch.tensordot(partitioned_grad, pre_conditioners_for_grad[i], dims=[[0], [0]])
+            i += 1
+
         return partitioned_grad
 
     def preconditioned_grad(self, grad: torch.Tensor) -> torch.Tensor:
@@ -344,74 +375,49 @@ class PreConditioner:
 
         merged_grad = self.partitioner.merge_partitions(pre_cond_partitioned_grads)
 
-        return torch.reshape(merged_grad, self.original_shape)
+        return merged_grad.reshape(self.original_shape)
+
+
+def build_graft(p: torch.Tensor, graft_type: int, diagonal_eps: float = 1e-10):
+    r"""Build Graft by given graft_type."""
+    if graft_type == LayerWiseGrafting.ADAGRAD:
+        return AdaGradGraft(p, diagonal_eps)
+    if graft_type == LayerWiseGrafting.RMSPROP:
+        return RMSPropGraft(p, diagonal_eps)
+    if graft_type == LayerWiseGrafting.SGD:
+        return SGDGraft(p)
+    if graft_type == LayerWiseGrafting.SQRTN:
+        return SQRTNGraft(p)
+    return Graft(p)
 
 
 @torch.no_grad()
-def power_iter(mat_g: torch.Tensor, error_tolerance: float = 1e-6, num_iters: int = 100) -> torch.Tensor:
-    r"""Power iteration.
+def power_iteration(mat_g: torch.Tensor, num_iters: int = 100) -> torch.Tensor:
+    r"""Compute the maximum eigenvalue of matrix, for scaling.
 
-        Compute the maximum eigenvalue of mat, for scaling. v is a random (uniform) vector with values in (-1, 1).
+        Mostly, power_iteration method is faster than torch.einval in case of the symmetric PSD matrix.
+        Also, I removed the validation, error of singular value every iteration, so that boosting the speed.
 
     :param mat_g: torch.Tensor. the symmetric PSD matrix.
-    :param error_tolerance: float. Iterative exit condition.
     :param num_iters: int. Number of iterations.
     """
-    v: torch.Tensor = 2.0 * torch.rand(list(mat_g.shape)[0], dtype=mat_g.dtype, device=mat_g.device) - 1
+    v = torch.randn(mat_g.shape[0], dtype=mat_g.dtype, device=mat_g.device)
+    mat_v = torch.empty_like(v)
 
-    error: Union[torch.Tensor, float] = 1.0
-    iters: int = 0
-    singular_val: Union[torch.Tensor, float] = 0.0
-    while error > error_tolerance and iters < num_iters:
-        v.div_(v.norm())
-        mat_v = torch.mv(mat_g, v)
-        s_v = torch.dot(v, mat_v)
+    for _ in range(num_iters):
+        torch.mv(mat_g, v, out=mat_v)
+        v.copy_(mat_v)
+        v.div_(torch.linalg.norm(v))
 
-        error = torch.abs(s_v - singular_val)
-
-        v = mat_v
-        singular_val = s_v
-        iters += 1
-
-    return singular_val
-
-
-@torch.no_grad()
-def matrix_power(mat_m: torch.Tensor, p: int) -> torch.Tensor:
-    r"""Compute mat_m^{p}.
-
-        Unroll the loop due to the performance.
-
-    :param mat_m: torch.Tensor. a square matrix.
-    :param p: int. a positive integer. usually p = (1, 2, 4, 8).
-    """
-    exponent: int = int(np.log2(p))
-
-    if exponent == 0:
-        return mat_m
-
-    mat_p2 = torch.matmul(mat_m, mat_m).float()
-    if exponent == 1:
-        return mat_p2
-
-    mat_p4 = torch.matmul(mat_p2, mat_p2).float()
-    if exponent == 2:
-        return mat_p4
-
-    mat_p8 = torch.matmul(mat_p4, mat_p4).float()
-    if exponent == 3:
-        return mat_p8
-
-    # most of the cases, never reached here.
-    return torch.matmul(mat_p8, mat_p8).float()
+    return (v.t() @ mat_g @ v).clamp_min_(1e-16)
 
 
 @torch.no_grad()
 def compute_power_schur_newton(
     mat_g: torch.Tensor,
     p: int,
-    iter_count: int = 100,
-    error_tolerance: float = 1e-6,
+    max_iters: int = 100,
+    error_tolerance: float = 1e-3,
     ridge_epsilon: float = 1e-6,
     max_error_ratio: float = 1.2,
 ) -> torch.Tensor:
@@ -433,90 +439,127 @@ def compute_power_schur_newton(
 
     :param mat_g: torch.Tensor. A square positive semi-definite matrix.
     :param p: int. a positive integer.
-    :param iter_count: int. Stop iterating after this many rounds.
+    :param max_iters: int. Stop iterating after this many rounds.
     :param error_tolerance: float. Threshold for stopping iteration.
     :param ridge_epsilon: float. We add this times I to G, to make is positive definite.
         For scaling, we multiply it by the largest eigenvalue of G.
     :param max_error_ratio: float. Sometimes error increases after an iteration before decreasing and converging.
         1.2 factor is used to bound the maximal allowed increase.
     """
-    shape: List[int] = list(mat_g.shape)
+    shape: List[int] = mat_g.shape
     if len(shape) == 1:
         return torch.pow(mat_g + ridge_epsilon, -1.0 / p)
 
-    identity = torch.eye(shape[0], device=mat_g.device, dtype=torch.float32)
+    identity = torch.eye(shape[0], dtype=mat_g.dtype, device=mat_g.device)
     if shape[0] == 1:
         return identity
 
-    max_ev = power_iter(mat_g).clamp_(min=1e-16)
-    ridge_epsilon *= max_ev
-    mat_g += ridge_epsilon * identity
+    mat_g += power_iteration(mat_g) * identity * ridge_epsilon
 
-    z = (1 + p) / (2 * torch.norm(mat_g))
+    z = (1 + p) / (2 * torch.linalg.norm(mat_g))
 
     mat_root = identity * torch.pow(z, 1.0 / p)
+
     mat_m = mat_g * z
 
     alpha: float = -1.0 / p
-    alpha_identity: torch.Tensor = (1.0 - alpha) * identity
-    error = torch.max(torch.abs(mat_m - identity))
-    count: int = 0
-    while error > error_tolerance and count < iter_count:
-        mat_m_i = alpha_identity + alpha * mat_m
-        new_mat_root = torch.matmul(mat_root, mat_m_i).float()
-        mat_m = torch.matmul(matrix_power(mat_m_i, p), mat_m).float()
+    alpha_identity = (1.0 - alpha) * identity
 
-        new_error = torch.max(torch.abs(mat_m - identity))
-        if new_error > error * max_error_ratio:
+    prev_error = torch.max(torch.abs(mat_m - identity))
+
+    for _ in range(max_iters):
+        mat_m_i = alpha_identity + alpha * mat_m
+
+        new_mat_root = torch.matmul(mat_root, mat_m_i)
+        torch.matmul(torch.linalg.matrix_power(mat_m_i, p), mat_m, out=mat_m)
+
+        error = torch.max(torch.abs(mat_m - identity))
+
+        # NOTE
+        # this is the main bottleneck that makes Scalable Shampoo slow.
+        # because it is handled on the Python side so values need to be on the CPU
+        # while XLA devices (e.g. TPU) doesn't seem to be affected.
+        if torch.logical_or(error > prev_error * max_error_ratio, error <= error_tolerance):
             break
 
         mat_root = new_mat_root
-        error = new_error
-        count += 1
+        prev_error = error
 
     return mat_root
 
 
 @torch.no_grad()
 def compute_power_svd(matrix: torch.Tensor, power: float) -> torch.Tensor:
-    r"""Compute G^{-1/p} using a SVD.
+    r"""Compute G^{-1/p} using SVD.
 
         Calculate SVD on the GPU. Sometimes, SVD on the CPU is faster than GPU, but based on the several experiments,
         CUDA seems much faster than on CPU.
 
     :param matrix: torch.Tensor. a square positive semi-definite matrix.
-    :param power: float. -1.0 / order.
+    :param power: float. rank.
     """
     u, s, vh = torch.linalg.svd(matrix, full_matrices=False)
-    s.pow_(power)
+    s.pow_(-1.0 / power)
     return u @ (s.diag() if len(matrix.shape) == 2 else s.diag_embed()) @ vh
 
 
-def merge_small_dims(shape_to_merge: List[int], max_dim: int) -> List[int]:
+def merge_small_dims(shape_to_merge: Union[List[int], torch.Size], max_dim: int) -> List[int]:
     r"""Merge small dimensions.
 
         If there are some small dimensions, we collapse them
             e.g. [1, 2, 512, 1, 2048, 1, 3, 4] --> [1024, 2048, 12] if max_dim = 1024
             [1, 2, 768, 1, 2048] --> [2, 768, 2048].
 
-    :param shape_to_merge: List. Shape to merge small dimensions.
+    :param shape_to_merge: Union[List[int], torch.Size]. Shape to merge small dimensions.
     :param max_dim: int. Maximal dimension of output shape used in merging.
     """
-    if shape_to_merge and np.all(np.array(shape_to_merge) == 1):
-        return [1]
-
-    resulting_shape: List[int] = []
+    merged_shape: List[int] = []
 
     product: int = 1
-    for d in shape_to_merge:
-        if product * d <= max_dim:
-            product *= d
-        else:
-            if product > 1:
-                resulting_shape.append(product)
-            product = d
+    for dim in shape_to_merge:
+        product *= dim
+        if product > max_dim:
+            merged_shape.append(product // dim)
+            product = dim
 
-    if product > 1:
-        resulting_shape.append(product)
+    merged_shape.append(product)
 
-    return resulting_shape
+    return merged_shape if len(merged_shape) > 1 else [1]
+
+
+def zero_power_via_newton_schulz_5(
+    g: torch.Tensor, num_steps: int = 10, eps: float = 1e-7, weights: Tuple[int, int, int] = (3.4445, -4.7750, 2.0315)
+) -> torch.Tensor:
+    r"""Compute the zeroth power / orthogonalization of G.
+
+    Newton-Schulz iteration to compute the zeroth power / orthogonalization of G. We opt to use a quintic iteration
+    whose coefficients are selected to maximize the slope at zero. For the purpose of minimizing steps, it turns out
+    to be empirically effective to keep increasing the slope at zero even beyond the point where the iteration no
+    longer converges all the way to one everywhere on the interval. This iteration therefore does not produce UV^T but
+    rather something like US'V^T where S' is diagonal with S_{ii}' ~ Uniform(0.5, 1.5), which turns out not to hurt
+    model performance at all relative to UV^T, where USV^T = G is the SVD.
+
+    :param g: torch.Tensor. matrix.
+    :param num_steps: int. number of iterations.
+    :param eps: float. add this times I to G, to make is positive definite. For scaling, we multiply it by the largest
+        eigenvalue of G.
+    :param weights: Tuple[int, int, int]. weights.
+    """
+    if len(g.shape) != 2:
+        raise ValueError('shape of g must be 2-dimensional')
+
+    x = g.bfloat16()
+    x.div_(x.norm().add_(eps))
+
+    if g.size(0) > g.size(1):
+        x = x.T
+
+    for _ in range(num_steps):
+        a = x @ x.T
+        b = weights[1] * a + weights[2] * a @ a
+        x = weights[0] * x + b @ x
+
+    if g.size(0) > g.size(1):
+        x = x.T
+
+    return x
